// Copyright 2021 Layotto Authors
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package lock

import (
	"os"
	"testing"

	"github.com/stretchr/testify/require"
)

const key = "lock-key-1234567"

func TestMain(m *testing.M) {
	SaveLockConfiguration("store1", map[string]string{strategyKey: strategyNone})
	SaveLockConfiguration("store2", map[string]string{strategyKey: strategyAppid})
	SaveLockConfiguration("store3", map[string]string{strategyKey: strategyDefault})
	SaveLockConfiguration("store4", map[string]string{strategyKey: strategyStoreName})
	SaveLockConfiguration("store5", map[string]string{strategyKey: "other-fixed-prefix"})
	// if strategyKey not set
	SaveLockConfiguration("store6", map[string]string{})
	os.Exit(m.Run())
}

func TestSaveLockConfiguration(t *testing.T) {
	testIllegalKeys := []struct {
		storename string
		prefix    string
	}{
		{
			storename: "lockstore01",
			prefix:    "a||b",
		},
	}
	for _, item := range testIllegalKeys {
		err := SaveLockConfiguration(item.storename, map[string]string{
			strategyKey: item.prefix,
		})
		require.NotNil(t, err)
	}
}

func TestGetModifiedLockKey(t *testing.T) {
	// use custom prefix key
	testIllegalKeys := []struct {
		storename string
		prefix    string
		key       string
	}{
		{
			storename: "lockstore01",
			prefix:    "a",
			key:       "c||d",
		},
	}
	for _, item := range testIllegalKeys {
		err := SaveLockConfiguration(item.storename, map[string]string{
			strategyKey: item.prefix,
		})
		require.Nil(t, err)
		_, err = GetModifiedLockKey(item.key, item.storename, "")
		require.NotNil(t, err)
	}
}

func TestNonePrefix(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store1", "appid1")
	require.Equal(t, "lock|||"+key, modifiedLockKey)
}

func TestAppidPrefix(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store2", "appid1")
	require.Equal(t, "lock|||appid1||lock-key-1234567", modifiedLockKey)
}

func TestAppidPrefix_WithEnptyAppid(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store2", "")
	require.Equal(t, "lock|||lock-key-1234567", modifiedLockKey)
}

func TestDefaultPrefix(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store3", "appid1")
	require.Equal(t, "lock|||appid1||lock-key-1234567", modifiedLockKey)
}

func TestStoreNamePrefix(t *testing.T) {
	key := "lock-key-1234567"

	modifiedLockKey, _ := GetModifiedLockKey(key, "store4", "appid1")
	require.Equal(t, "lock|||store4||lock-key-1234567", modifiedLockKey)
}

func TestOtherFixedPrefix(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store5", "appid1")
	require.Equal(t, "lock|||other-fixed-prefix||lock-key-1234567", modifiedLockKey)
}

func TestLegacyPrefix(t *testing.T) {
	modifiedLockKey, _ := GetModifiedLockKey(key, "store6", "appid1")
	require.Equal(t, "lock|||appid1||lock-key-1234567", modifiedLockKey)
}

func TestPrefix_StoreNotInitial(t *testing.T) {
	// no config for store999
	modifiedLockKey, _ := GetModifiedLockKey(key, "store999", "appid99")
	require.Equal(t, "lock|||appid99||lock-key-1234567", modifiedLockKey)
}
